/*
    M5Stackと心拍センサユニットで酸素飽和度と心拍数、心拍パルスを取得し可視化するプログラム
    [MQTT送信対応版]

    Copyright (C) 2021 SHIRAFU Makoto <shirafu@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    《概要》
    心拍センサから取得した値をM5StackのLCD画面上に表示するのと合わせて、指定された
    MQTTサーバに送信します。
    別に提供するモニタプログラムを使うことにより複数台の心拍センサーからの計測値を1台以
    上のパソコンでモニタリングすることができます。

    【重要】このプログラムを使用するためにはActiveMQ( https://activemq.apache.org )な
    どMQTT環境の構築と設定が必要です。
    MQTTサーバで設定したユーザ名/パスワード、トピック名を本プログラムに適用したのちビルドし
    M5Stack環境にインストールします。


    ■必要なハードウェア
    ・M5Stack シリーズ
      ESP32 Basic Core IoT Development Kit
      https://shop.m5stack.com/collections/stack-series/products/basic-core-iot-development-kit

    ・M5Stack用心拍センサユニット
      Mini Heart Rate Unit (MAX30100) Pulse Oximeter
      https://shop.m5stack.com/products/mini-heart-unit

    ■必要なソフトウェア
    ・Visual Studio Code
      https://code.visualstudio.com

    ・PlatformIO
      https://platformio.org/install/ide?install=vscode

    ■使用しているArduinoライブラリ
    ・M5Stack Library
      https://github.com/m5stack/M5Stack

    ・Arduino-MAX30100
      https://github.com/oxullo/Arduino-MAX30100

    ■ソースの構成
      ・main.cpp                    ...   可視化するプログラム本体
      ・MAX30100_PulseOximeter.cpp  ...   Arduino-MAX30100より心拍パルスをとれるように修正
      ・MAX30100_PulseOximeter.h    ...   Arduino-MAX30100より心拍パルスをとれるように修正

  */

#include <Arduino.h>
#include <M5Stack.h>
#include <Wire.h>
#include "MAX30100_PulseOximeter.h"

#include <WiFi.h>
#include <WiFiMulti.h>
#include <HTTPClient.h>
#include <PubSubClient.h>
#include <time.h>
#include <TimeLib.h>

// 酸素飽和度と心拍数の更新間隔(ms)
#define STATUS_INTERVAL_MS      1000
// 心拍グラフの更新間隔(ms)
#define GRAPH_INTERVAL_MS       15

// 心拍数を正常とする下限値
#define NORMAL_HEART_RATE_LOWER 60
// 心拍数を正常とする上限値
#define NORMAL_HEART_RATE_UPPER 100

// 液晶の横サイズ
#define LCD_WIDTH 320
// 液晶の縦サイズ
#define LCD_HEIGHT 240

// ステータス表示の横サイズ
#define IMG_WIDTH 320
// ステータス表示の縦サイズ
#define IMG_HEIGHT 140

//--------------------------------------------------------------------------
// ここから　稼働環境にあわせて調整してください
//--------------------------------------------------------------------------
// データの送受信に使用するWi-FiのSSID
const char *wifi_ssid = "《Wi-FiのSSID》";
// データの送受信に使用するWi-Fiの接続パスワード
const char *wifi_password = "《Wi-Fiの接続パスワード》";

// MQTTサーバのIPアドレスまたはホスト名
const char* mqtt_host = "《MQTTサーバのIPアドレス》";
// MQTTサーバのポート
const int mqtt_port = 61616;       

// MQTTサーバの接続ユーザ名
const char *mqtt_username = "《MQTTサーバの接続ユーザ名》";
// MQTTサーバの接続パスワード
const char *mqtt_password = "《MQTTサーバの接続パスワード》";
// 送信先のMQTTトピック名
const char* topic = "《MQTTサーバ上の送信先トピック名》";

// このM5StackのデバイスID(モニタプログラムでフィルタリングに使用)
const char *device_id = "《インストール先のM5Stackの識別デバイスID》";

//--------------------------------------------------------------------------
// ここまで　稼働環境にあわせて調整してください
//--------------------------------------------------------------------------

// NTPサーバ
const char *ntp_server = "ntp.jst.mfeed.ad.jp";
// 時差(NTP)
const long gmt_offset_sec = 9 * 3600;
// 夏時間(NTP)
const int daylight_offset_sec = 0;

// パルスオキシメータ
PulseOximeter pox;

// 酸素飽和度
uint8_t spo2 = 0;
// 心拍数
uint8_t hr = 0;
// 心拍パルス値
float pulse;

// 直前のステータス表示更新
uint32_t tsLastReport = 0;
// 直前のグラフ表示更新
uint32_t tsLastPulse = 0;

// 心拍グラフの描画位置(X)
int cx = 0;
// 心拍グラフの直前の描画位置(X)
int px = -1;
// 心拍グラフの直前の描画位置(Y)
int py = -1;

// ステータス表示のオフスクリーンバッファ
TFT_eSprite statusimg = TFT_eSprite(&M5.Lcd);

// Wi-Fi接続
WiFiClient wifi_client;
// MQTTクライアント
PubSubClient mqtt_client(wifi_client);

// MQTT送信データ
char mqtt_payload[1024 * 2];       

// 日付表示整形用
char date_text[30];
// 時刻表示整形用
char time_text[30];

//
//  SpO2値が妥当か評価する
//
int isValidSpo2(uint8_t spo2) {
  return spo2 > 0 && spo2 <= 100;
}

//
// ステータス(酸素飽和度と心拍数)の更新
//
void updateStatus() {

  // 背景色
  uint32_t bg;
  // 前景色
  uint32_t fg = TFT_WHITE;

  // SPO2の値から色を決定
  //  96 以上:  黒 / 白
  //  90 〜 95: 黄色 / 黒
  //  75 〜 89: ピンク / 黒
  //  74以下:    赤 / 白
  if (spo2 >= 96) {
    bg = TFT_BLACK;
  } else if (spo2 >= 90) {
    bg = TFT_YELLOW;
    fg = TFT_BLACK;
  } else if (spo2 >= 75) {
    bg = TFT_PINK;
    fg = TFT_BLACK;
  } else if (spo2 > 0) {
    bg = TFT_RED;
  } else {
    bg = TFT_BLACK;
  }

  // オフスクリーンバッファを背景色で塗りつぶして前景色を設定
  statusimg.fillRect(0, 0, IMG_WIDTH, IMG_HEIGHT, bg);
  statusimg.setTextColor(fg);

  // 見出しを描画
  statusimg.setCursor(0, 0);
  statusimg.setTextSize(3);
  statusimg.print("SPO2(%)");

  statusimg.setCursor(0, 110);
  statusimg.setTextSize(3);
  statusimg.print("PR(bpm)");

  // 酸素飽和度と心拍数を描画
  if (isValidSpo2(spo2)) {
    statusimg.setCursor(120, 40);
    statusimg.setTextSize(7);
    statusimg.print(spo2);

    statusimg.setCursor(160, 110);
    statusimg.setTextSize(3);
    statusimg.print(hr);

    // 心拍数の警告
    if (hr < NORMAL_HEART_RATE_LOWER || hr > NORMAL_HEART_RATE_UPPER) {
      statusimg.setCursor(250, 110);
      statusimg.print("(!)");
    }
  } else {
     // 酸素飽和度が無効の場合、酸素飽和度と心拍数を表示しない
    statusimg.setCursor(120, 40);
    statusimg.setTextSize(7);
    statusimg.print("--");

    statusimg.setCursor(160, 110);
    statusimg.setTextSize(3);
    statusimg.print("--");
  }

	// 現在日時を取得
	time_t timer;
	time(&timer);
	struct tm timeinfo;
	localtime_r(&timer, &timeinfo);

	// 日付と時刻を表示用に整形
	sprintf(date_text, "%04d/%02d/%02d", (1900 + timeinfo.tm_year), (1 + timeinfo.tm_mon), timeinfo.tm_mday);
	sprintf(time_text, "%02d:%02d:%02d", timeinfo.tm_hour, timeinfo.tm_min, timeinfo.tm_sec);

  statusimg.setTextSize(2);
  statusimg.drawString(date_text, 180, 0);
  statusimg.drawString(time_text, 180, 20);

  // ステータスの表示更新
	statusimg.pushSprite(0, 0, TFT_TRANSPARENT);
}

//
// 心拍グラフの更新
//
void updateGraph(float value) {

  // 以前の描画を消して
  M5.Lcd.drawLine(cx, 141, cx, LCD_HEIGHT, TFT_BLACK);

  // 心拍パルスの値を適当な位置に描画するよう算出(今回座標のY)
  int cy = 220 - (value + 50) / 5;
  if (cy <= 141) { cy = 141; }
  // 前回座標(X)がマイナス値の場合、前回座標(Y)を今回座標(Y)で上書き
  if (px < 0) { py = cy; }

  // 前回座標から今回座標まで新しい線を描画
  M5.Lcd.drawLine(px, py, cx, cy, CYAN);

  // 描画座標を待避
  px = cx++; py = cy;
  if (cx >= LCD_WIDTH) { cx = 0; px = 0; }
}

//
//  M5Stack上の現在日時の表示を更新
//
void updateCurrentDatetime() {

	// 現在日時を取得
	time_t timer;
	time(&timer);

	struct tm timeinfo2;
	localtime_r(&timer, &timeinfo2);

	sprintf(date_text, "%04d/%02d/%02d", (1900 + timeinfo2.tm_year), (1 + timeinfo2.tm_mon), timeinfo2.tm_mday);
	sprintf(time_text, "%02d:%02d:%02d", timeinfo2.tm_hour, timeinfo2.tm_min, timeinfo2.tm_sec);
}

//
// 酸素飽和度と心拍数を送信(MQTT)
//
void sendMQTT(uint8_t spo2, uint8_t hr) {
  // デバイスID 日付 時刻 SPO2=酸素飽和度,PR=心拍数
  sprintf(mqtt_payload, "%s\t%s\t%s\tSPO2=%d,PR=%d", device_id, date_text, time_text, spo2, hr);
  mqtt_client.publish(topic, mqtt_payload);
}

//
// 心拍パルスを送信(MQTT)
//
void sendMQTT(float pulse) {
  // デバイスID 日付 時刻 Pulse=心拍パルス値
  sprintf(mqtt_payload, "%s\t%s\t%s\tPulse=%f", device_id, date_text, time_text, pulse);
  mqtt_client.publish(topic, mqtt_payload);
}

//
// 心拍検知ハンドラ
//
void beatDetectedHandler()
{
  // 心拍検知したときにBEEP
  M5.Speaker.tone(1200, 10);

  // 心拍に合わせてMQTTへ送信
  sendMQTT(spo2, hr);

  updateStatus();
}

//
//  Wi-Fiに接続 
//
int connectWifi() {

    M5.Lcd.setTextSize(2);
    M5.Lcd.print("Connecting to Wi-Fi ");
    Serial.print("Connecting to Wi-Fi ");

    WiFi.begin(wifi_ssid, wifi_password);
    while (WiFi.status() != WL_CONNECTED)
    {
      delay(500);
      M5.Lcd.print(".");
      Serial.print(".");
    }
    M5.Lcd.print("\nConnected.!");
    Serial.print("\nConnected.!");

    return 0;
}

///
/// NTPに接続し日時設定の同期を行う
///
int syncNtp() {

  M5.Lcd.setTextSize(2);
  M5.Lcd.println("Adjusting the time with NTP Server");
  Serial.println("Adjusting the time with NTP Server");

	configTime(gmt_offset_sec, daylight_offset_sec, ntp_server);
	struct tm timeinfo;
	if (!getLocalTime(&timeinfo))
	{
		M5.Lcd.println("Failed to obtain time");
		Serial.println("Failed to obtain time");
		return -1;
	}
  M5.Lcd.println("Successful synchronization.");
  Serial.println("Successful synchronization.");

  return 0;
}

//
//  MQTTサーバへ接続する
//
void connectMqtt()
{
  M5.Lcd.setTextSize(2);
  M5.Lcd.println("Connecting to MQTT Server");
  Serial.println("Connecting to MQTT Server");

    mqtt_client.setServer(mqtt_host, mqtt_port);
    while( ! mqtt_client.connected() ) {
        // 識別用のIDを生成
        String clientId = "M5POX-" + String(random(0xffff), HEX);
        if ( mqtt_client.connect(clientId.c_str(), mqtt_username, mqtt_password) ) {
          M5.Lcd.print("\nConnected.!");
          Serial.print("\nConnected.!");
          continue;
        }

        delay(1000);
        randomSeed(micros());

        M5.Lcd.print(".");
        Serial.print(".");
    }
}

//
//  セットアップ
//ˇ
void setup()
{
  M5.begin();
	Wire.begin();
  Serial.begin(115200);

  // オフスクリーンバッファ初期化
	statusimg.setColorDepth(4);
	statusimg.createSprite(IMG_WIDTH, IMG_HEIGHT);
	statusimg.setSwapBytes(false);

  // Wi-Fi初期化
  connectWifi();
  // NTP初期化
  syncNtp();
  // MQTT接続
  connectMqtt();

  // センサーの初期化
  if (!pox.begin()) {
    M5.Speaker.beep();
    M5.Lcd.println("Sensor initialization failed");
    for(;;);
  }

  // センサーのLED設定(デフォルト値の50mAではうまくとれない)
  pox.setIRLedCurrent(MAX30100_LED_CURR_7_6MA);
  // センサーの心拍検知ハンドラ設定
  pox.setOnBeatDetectedCallback(beatDetectedHandler);

  M5.Lcd.drawLine(0, 140, LCD_WIDTH, 140, TFT_WHITE);
}

//
//  メインループ
//
void loop()
{
  // M5Stackとセンサーの状態を更新
  // (M5Stackの状態更新をしないとBEEPが止まらなくなる)
  m5.update();
  pox.update();

  uint32_t now = millis();

  // 前回更新から STATUS_INTERVAL_MS ミリ秒分経過していたらステータス更新
  if (now - tsLastReport > STATUS_INTERVAL_MS) {
    // 酸素飽和度と心拍数を取得
    hr = (uint8_t)pox.getHeartRate();
    int prev = spo2;
    spo2 = pox.getSpO2();
    // 前回有効な酸素飽和度が取得できて今回は無効値の場合、MQにデータを送信
    // (モニタ側の表示を現在値とするため)
    if (isValidSpo2(prev) && !isValidSpo2(spo2)) {
      sendMQTT(spo2, hr);
    }
    updateCurrentDatetime();
    updateStatus();
    tsLastReport = now;
  }

  // 前回更新から GRAPH_INTERVAL_MS ミリ秒分経過していたらグラフ更新
  if (now - tsLastPulse > GRAPH_INTERVAL_MS) {
    // 心拍パルスを取得
    pulse = pox.getFilteredPulseValue();
    updateGraph(pulse);
    sendMQTT(pulse);
    tsLastPulse = now;
  }

  mqtt_client.loop();
}
